import * as d3 from 'd3';
import { globalSetting } from '../global';
import { NetlistRender } from '.';

export class CellRender {
    /**
     * 
     * @param {d3.Selection} selection 
     * @param {NetlistRender} rootRender 
     */
    constructor(selection, rootRender) {
        this.parentSelection = selection;
        this.rootRender = rootRender;

        /**
         * @type {BasicD3DataItem[]}
         */
        this.data = [];

        /**
         * @description id 到管理数据项的映射
         * @type {Map<string, BasicD3ManagmentItem[]>}
         */
        this.id2manager = rootRender.id2manager;
    }

    /**
     * @description 将 elknode 关于 器件 的数据添加为 d3 数据项目
     * @param {ElkNode} node 
     * @param {HTMLElement} element
     */
    addAsD3DataItem(node, element) {
        this.data.push({
            id: node.id,
            element,
            x: node.x,
            y: node.y,
            width: node.width,
            height: node.height,
            fill: 'var(--main-dark-color)'
        });
    }

    render() {
        const data = this.data;
        const rootRender = this.rootRender;
        const id2manager = this.id2manager;
        const _this = this;

        let cellSelections = this.parentSelection.selectAll('svg')
            .data(data)
            .enter()
            .append(data => {
                const element = data.element;
                element.setAttribute('x', data.x);
                element.setAttribute('y', data.y);
                if (globalSetting.renderAnimation) {
                    element.setAttribute('stroke-opacity', 0);
                }
                return element;
            });
        
        if (globalSetting.renderAnimation) {
            cellSelections = cellSelections
                .transition()
                .duration(1000)
                .attr('stroke-opacity', 1)
                .attr('class', 'grab')
                .on('end', function (data) {                    
                    const cellSelection = d3.select(this);
                    const manager = _this.createDataManager(cellSelection, data);
                    registerDragEvent(manager, rootRender);
                });
        } else {
            cellSelections = cellSelections
                .attr('class', 'grab')
                .each(function (data) {
                    const cellSelection = d3.select(this);
                    const manager = _this.createDataManager(cellSelection, data);
                    registerDragEvent(manager, rootRender);
                });
        }

        this.selections = cellSelections;
        return cellSelections;
    }

    /**
     * 
     * @param {d3.Selection} selection 
     * @param {BasicD3DataItem} data 
     * @returns {BasicD3ManagmentItem}
     */
    createDataManager(selection, data) {
        const id2manager = this.id2manager;
        // 创建拖拽上下文
        const dragContext = {
            elkGraph: {
                // elk 是无状态的，id 取什么名字都行
                id: 'root',
                children: [],
                edges: [],
                layoutOptions: {}
            }
        }
        const managerItem = {
            data,
            selection,
            type: 'cell',
            dragContext
        };

        if (!id2manager.has(data.id)) {
            id2manager.set(data.id, []);
        }

        id2manager.set(data.id, managerItem);
        return managerItem;
    }
}

/**
 * @description 注册关于 器件 的拖动事件
 * 
 * 需要提取最小拓扑子图，然后重新调整各个区域的尺寸
 * @param {BasicD3ManagmentItem} manager
 * @param {NetlistRender} rootRender
 */
export function registerDragEvent(manager, rootRender) {
    // 创建拖拽行为
    const drag = d3.drag();

    drag.on('start', async event => dragStart(event, manager, rootRender));
    drag.on('drag', async event => dragged(event, manager, rootRender));
    drag.on('end', async event => dragEnd(event, manager, rootRender))
    
    manager.selection.call(drag);
}

/**
 * 
 * @param {d3.D3DragEvent} event 
 * @param {BasicD3ManagmentItem} manager
 * @param {NetlistRender} rootRender 
 */
function dragStart(event, manager, rootRender) {
    const selection = manager.selection;
    selection.attr('class', 'grabbing');
    
    // 更新拓扑图中各个节点的节点坐标


}

/**
 * 
 * @param {d3.D3DragEvent} event 
 * @param {BasicD3ManagmentItem} manager
 * @param {NetlistRender} rootRender 
 */
function dragged(event, manager, rootRender) {
    // 当拖动结束时，D3 会根据绑定的数据（data.x 和 data.y）重新渲染元素，导致元素回到初始位置。
    // 所以需要 手动更新 data.x 和 data.y
    const selection = manager.selection;
    const data = manager.data;

    data.x = event.x;
    data.y = event.y;
    
    selection
        .attr('x', event.x)
        .attr('y', event.y);
    
    // 根据最小拓扑图，提取出关键点，重新计算布局
    const context = manager.dragContext;
    console.log(context.elkGraph);
}

/**
 * 
 * @param {d3.D3DragEvent} event 
 * @param {BasicD3ManagmentItem} manager
 * @param {NetlistRender} rootRender
 */
function dragEnd(event, manager, rootRender) {
    const selection = manager.selection;
    selection.attr('class', 'grab');
}